#ifndef HOUGHPOINT_H
#define HOUGHPOINT_H

#include "TVector2.h"

/********************************************************************
 * The HoughPoint class represents the point of intersection of     *
 * two trajectories in in hough space.  The class contains the two  *
 * data points which are transformed into trajectories as well as   *
 * the resulting hough parameters.  This allows for easy mapping    *
 * between the accumulation in the hough space and the data points  *
 * back in the normal data space.                                   *
 *                                                                  *
 * This version is optimized for a 2D linear hough transform.       *
 *                                                                  *
 * anorman@fnal.gov                                                 *
 ********************************************************************/

namespace novaddt {

  class DAQHit;

  class HoughPoint;
}

// Notes and suggestions:
//
// Holding a shared pointer to a HoughData was a false economy: a
// HoughPoint below is two pointers, and two floats: a shared pointer as
// before was multiple contained classes and lots of reference-counting
// machinery.
//
// Use compiler-default methods whenever possible. Here, two hits are
// used rather than an array of hits to allow this.
//
// reset() is usually the term used for what you had as clear(). clear()
// usually refers to containers.
//
// size() is similarly used by containers and could confuse someone who
// expected the answer 2 (two hits). If someone wants sizeof(), they'll
// use sizeof().
//
// I haven't changed it, but rather than print(), a more versatile idiom
// is to implement a free operator << () which can be used with any
// stream. This often calls a member function with ostream
// &print(ostream &, <type>).
//
// calc() is void: no point in returning something you're not going to
// use.
//
// By taking calc() private, having reset(...) instead of setPoint() and
// relying on compiler-generated copy/asignment, we now have a class
// invariant: _cosv and _doca will always be correct for the given
// (_p1, _p2).
class novaddt::HoughPoint {
public:
  HoughPoint();
  HoughPoint(DAQHit const & p1, TVector2 const & coord1, DAQHit const & p2, TVector2 const & coord2);
  // Use default copy/move constructors and assignment.

  /// Initialization of Data members
  void reset(DAQHit const & p1, TVector2 const & coord1, DAQHit const & p2, TVector2 const & coord2);
  void reset();

  void print() const;

  float const & getCosv()  const { return _cosv; };
  float const & getDoca()  const { return _doca; };

  float const & getDist()  const { return _dist; };
  float const & getVpro()  const { return _vpro; };

  DAQHit const & getCh1()  const { return *_p1;  };
  DAQHit const & getCh2()  const { return *_p2;  };  

private:
  void calc();

  DAQHit const *   _p1; // Non-owning: consider cet::exempt_ptr.
  TVector2 const * _coord1;
  DAQHit const *   _p2; // Non-owning: consider cet::exempt_ptr.
  TVector2 const * _coord2;

  // Computed data members
  float _cosv;
  float _doca;

  float _dist;
  float _vpro;

};

#endif /* DAQHIT_H */
