////////////////////////////////////////////////////////////////////////
// Class:       MoonShadow
// Module Type: filter
// \file:       MoonShadow_module.cc
// \brief   	Select tracks originating from the direction of the moon or sun
// \author 	Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindOneP.h"

#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/Track3D.h"
#include "DDTBaseDataProducts/EventHeader.h"

// To convert TDC to NOvA time
#include <NovaTimingUtilities/TimingUtilities.h>
// To convert vectors to lenght
#include "SlowMonopoleTrigger/Constants.h"

// This will ultimately come from an external UPS product
#include "MoonShadowTrigger/NOVASLocate.h"

// 
#include <TH2D.h>
#include <iostream>
#include <stdlib.h>
#include <fstream>

namespace novaddt
{
  class MoonShadow: public art::EDFilter
  {
  public:
    explicit MoonShadow(const fhicl::ParameterSet& p);
    virtual ~MoonShadow();

    virtual bool filter(art::Event & e);

    virtual void beginJob();
    virtual void endJob();

  // Why is it protected
  protected:
    std::string FindEphemerisFile() const;
    std::string FindSmartPrescaleLookup() const;
    
    // Angle manipulation
    double inline RadToDeg(double rad) const;
    double inline DegToRad(double deg) const;
    
    // Vector manipulation
    TVector3 AnglesToVector(double zen, double azi) const;
    double GetDot(double zen1, double azi1, double zen2, double azi2) const;
    TVector3 FarDetToAstron(TVector3& fdVector) const;
    TVector3 AstronToFarDet(TVector3& astrVector) const;
    double GetPhysLength(TVector3 vec) const;
    double GetEffectiveArea();
    void InitLookupTable();
    double GetAngularPrescale(double zen1, double azi1, double zen2, double azi2);
    bool GetAngularPrescaleDecision(double zen1, double azi1, double zen2, double azi2);
    double GetRate(double zen, double azi);
    double GetSmartPrescale(double zen1, double azi1, double zen2, double azi2);
    double GetSmartPrescale(double zen, double azi);
    bool GetSmartPrescaleDecision(double prescale);
    void GetHisto(std::string filename, int xmax, int ymax, TH2D* hist);
    
    // Time
    time_t GetTime(art::Event& e);

    int fPrescale;

    std::string fTrack3DLabel;
    std::string fInputHitLabel;
    std::string fInputHitInstance;
    double fDotCut;
    double fLengthCut;
    int fBinCut;

    shadow::NOVASLocate fLocate;

    int fNEvents, fNTracks, fNSunTrigs, fNMoonTrigs;
    
    // Lookup table for introducing "smart prescale"
    TH2D* rate_collected_angular;
    int bin_size;
    
    // Limit on what rate is safe for detector hardware
    double max_rate;
    // We will cut more tracks on the border of 10 degree window than in the middle.
    double angular_stdev;
    // file name of lookup table
    std::string fLookupFile;
    
  };


  MoonShadow::MoonShadow(const fhicl::ParameterSet& pset)
    : fPrescale(pset.get<int>("prescale")),
      fTrack3DLabel(pset.get<std::string>("Track3DLabel")),
      fInputHitLabel(pset.get<std::string>("InputHitLabel")),
      fInputHitInstance(pset.get<std::string>("InputHitInstance")), 
      fDotCut(cos(5*M_PI/180) /*cos(pset.get<double>("AngleCut")*M_PI/180)*/),
      fLengthCut(pset.get<double>("LengthCut")),
      fBinCut(pset.get<int>("BinCut")),
      fLocate(FindEphemerisFile()),
      fLookupFile(FindSmartPrescaleLookup()),
      max_rate(pset.get<double>("AvgRate")),
      angular_stdev(pset.get<double>("AngularSuppression")),
      fNEvents(0), fNTracks(0), fNSunTrigs(0), fNMoonTrigs(0)
  {
    produces< std::vector<TriggerDecision>>();
    InitLookupTable();
    srand(time(NULL));
    
    bin_size = 1;
    
  }
  
  double MoonShadow::GetAngularPrescale(double zen1, double azi1, double zen2, double azi2)
  {
      double dTheta = acos(GetDot(zen1, azi1, zen2, azi2)) * 180 / M_PI;
      return exp(-dTheta*dTheta/(angular_stdev*angular_stdev));
  }
  bool MoonShadow::GetAngularPrescaleDecision(double zen1, double azi1, double zen2, double azi2)
  {
      double prescale = GetAngularPrescale(zen1, azi1, zen2, azi2);
      
      double r = (double)rand() / (double)RAND_MAX;
      
      return r < prescale;
  }

void MoonShadow::GetHisto(std::string filename, int xmax, int ymax, TH2D* hist)
{
  ifstream data;
  data.open(filename.c_str());
  for(int j=0; j<ymax;j++)
  {
    for(int i=0; i<xmax; i++)
    {
      double val;
      data >> val;
      hist->SetBinContent(i,j, val);
    }
  }

  data.close();
}

  
  void MoonShadow::InitLookupTable()
  {
     // Init lookup table and make 2d hist with expected rate
     // Get histogram with full rates in 2degx2deg bins
     int azi_max = 360;
     int zen_max = 90;
     
     rate_collected_angular = new TH2D("LookupPrescale", "", azi_max, 0, 360, zen_max, 0, 90);
     
     GetHisto(fLookupFile, azi_max, zen_max, rate_collected_angular);
  }
  // Prescale factor taken for sum of rates in 2 positions
  double MoonShadow::GetRate(double zen, double azi)
  {
  	if(zen > 90)
		return 0;
	// Smart prescale is ratio of maximal rate to expected rate divided by angular prescale factor for given position
	int bin_azi = round(azi/bin_size);
	int bin_zen = round(zen/bin_size);
	
	//return 0;
	return rate_collected_angular->GetBinContent(bin_azi, bin_zen);
  }
  double MoonShadow::GetSmartPrescale(double zen1, double azi1, double zen2, double azi2)
  {
	double rate1 = GetRate(zen1, azi1);
	double rate2 = GetRate(zen2, azi2);
		
	if(rate1 == 0 && rate2 == 0)
		return 0; 

	return max_rate / (rate1 + rate2);
  }
  double MoonShadow::GetSmartPrescale(double zen, double azi)
  {
  	double rate = GetRate(zen, azi);
	
	if(rate == 0)
		return 0;
	
	return max_rate / rate;
  }
  bool MoonShadow::GetSmartPrescaleDecision(double prescale)
  {
  	// Is it good enough for small prescale factors ~1/100?
	double r = (double)rand() / (double)RAND_MAX;
	return r < prescale;
  }

  MoonShadow::~MoonShadow()
  {
  }

  void MoonShadow::beginJob()
  {
  }
  
  // Conversion radians <-> degrees
  double inline MoonShadow::DegToRad(double deg) const
  {
    return deg * M_PI / 180;
  }
  double inline MoonShadow::RadToDeg(double rad) const
  {
    return rad * 180 / M_PI;
  }
  
  // ???
  std::string MoonShadow::FindEphemerisFile() const
  {
    cet::search_path sp("FW_SEARCH_PATH");
    std::string ephemerisFname;
    sp.find_file(("MoonShadowTrigger/novas/JPLEPH"), ephemerisFname);
    assert(!ephemerisFname.empty());
    return ephemerisFname;
  }
  // find lookup table file
  std::string MoonShadow::FindSmartPrescaleLookup() const
  {
    cet::search_path sp("FW_SEARCH_PATH");
    std::string lookupFname;
    sp.find_file(("MoonShadowTrigger/smart_prescale/PrescaleLookup.txt"), lookupFname);
    assert(!lookupFname.empty());
    return lookupFname;
  }


  // Creates unit vector pointing to a point on unit sphere with theta=zen, phi = azi
  TVector3 MoonShadow::AnglesToVector(double zen, double azi) const
  {
    zen *= M_PI / 180;
    azi *= M_PI / 180;
    // Doesn't matter which axis is which
    return TVector3(sin(zen)*cos(azi), sin(zen)*sin(azi), cos(zen));
  }
  // Gets dot product of unit vectors with given angles
  double MoonShadow::GetDot(double zen1, double azi1, double zen2, double azi2) const
  {
    if(zen1 == zen2 && azi1 == azi2)
      return 1;

    zen1 *= M_PI / 180;
    zen2 *= M_PI / 180;
    azi1 *= M_PI / 180;
    azi2 *= M_PI / 180;
    double dot = sin(zen1)*cos(azi1)*sin(zen2)*cos(azi2) + sin(zen1)*sin(azi1)*sin(zen2)*sin(azi2) + cos(zen1)*cos(zen2);
    
    return dot;
  }
  
  // Transfoms vector in FarDet coordinate system to vector in astronomical coordinate system
  TVector3 MoonShadow::FarDetToAstron(TVector3& fdVector) const
  {
    throw "Not implemented";
    
  }
  // Transfoms vector in astronomical coordinate system to vector in FarDet coordinate system
  TVector3 MoonShadow::AstronToFarDet(TVector3& fdVector) const
  {
    throw "Not implemented";
    
  }
  // Get length of the vector in cm
  double MoonShadow::GetPhysLength(TVector3 vec) const
  {
    double x = vec.X() * novaddt::smt::Constants::CELL_WIDTH;
    double y = vec.Y() * novaddt::smt::Constants::CELL_WIDTH;
    double z = vec.Z() * novaddt::smt::Constants::PLANE_WIDTH;
    
    return sqrt(x*x + y*y + z*z);
  }
  
  
  // Gets absolute time of the event. This function shoudl be used to determine time up to uncertainty bigger than hit time
  // difference
  time_t MoonShadow::GetTime(art::Event& e)
  { 
    art::Handle<novaddt::EventHeader> header;

    e.getByLabel(fInputHitLabel, fInputHitInstance, header);
    
    uint64_t secs = header->timeStart/novadaq::timeutils::NOVA_TIME_FACTOR;

    uint64_t nova_time = novadaq::timeutils::NOVA_EPOCH + secs;
    
    return time_t(nova_time);
  }
  
  
  bool MoonShadow::filter(art::Event& evt)
  {
    ++fNEvents;

    std::unique_ptr<std::vector<TriggerDecision>> td(new std::vector<TriggerDecision>);

    // Grab the tracks from the event
    art::Handle<std::vector<Track3D>> tracks;
    evt.getByLabel(fTrack3DLabel, tracks);

    // We need the hitlist just so we can get the times for the trigger
    art::FindOneP<HitList> hls(tracks, evt, fTrack3DLabel);
    
    //time_t t = GetTime(hls);
    time_t t = GetTime(evt);
    // std::cout << "Time_t: " << asctime(gmtime(&t)) << std::endl;
    
    double zen_moon, azi_moon;
    fLocate.GetMoonPosition(t, zen_moon, azi_moon);
    double zen_sun, azi_sun;
    fLocate.GetSunPosition(t, zen_sun, azi_sun); 
    
    int trackIndex = 0;
    
    double smartPrescale = GetSmartPrescale(zen_moon, azi_moon, zen_sun, azi_sun);
    double smartPrescaleSun = GetSmartPrescale(zen_sun, azi_sun);
    double smartPrescaleMoon = GetSmartPrescale(zen_moon, azi_moon);
    double smartPrescaleTot = smartPrescaleSun + smartPrescaleMoon;
    smartPrescaleSun = smartPrescale * (smartPrescaleSun/smartPrescaleTot);
    smartPrescaleMoon = smartPrescale * (smartPrescaleMoon/smartPrescaleTot);

    for(const Track3D& track: *tracks){
      const HitList& hlist = *hls.at(trackIndex);
      ++trackIndex;
    
      TVector3 start = track.Start();
      TVector3 end = track.End();
      if(start.Y() < end.Y()) std::swap(start, end);
      
      // WORK
      //Check whether 3D track vector makes sense
      if(
      (start.X() == 0 && end.X() == 0) || 
      (start.Y() == 0 && end.Y() == 0) || 
      (start.Z() == 0 && end.Z() == 0))
      {
        continue;
      }      
      
      
      
      // Apply length cuts
      if(GetPhysLength(start-end) < fLengthCut || hlist.size() < fBinCut)
      {
        continue;
      }

      
      ++fNTracks;


      


      // Point back along the track towards the moon
      const TVector3 dir = (start-end).Unit();

      const double zen_trk = acos(dir.Y()) * 180 / M_PI;
      // docdb 5485 / numix 17 says FD is oriented -27o51'26'' from true North  
      // azi_trk = atan2(-dir.X(), dir.Z()) * 180 / M_PI - 27.857222;      
      // Or: 332o03'58.071769" (clockwise from North). Email from Virgil Bocean
      // to Alec Habig.                                                        
      double azi_trk = atan2(-dir.X(), dir.Z()) * 180 / M_PI + 332.142778;
      if(azi_trk < 0) azi_trk += 360;
      if(azi_trk > 360) azi_trk -= 360;
      
      const double dotMoon = GetDot(zen_moon, azi_moon, zen_trk, azi_trk);
      const double dotSun = GetDot(zen_sun, azi_sun, zen_trk, azi_trk);

      // Inline &&'s call all these functions, so this is made to avoid extra calling
      bool trigMoon = (dotMoon > fDotCut);
      if(trigMoon) { trigMoon = trigMoon && GetSmartPrescaleDecision(smartPrescale); }
      if(trigMoon) { trigMoon = trigMoon && GetAngularPrescaleDecision(zen_moon, azi_moon, zen_trk, azi_trk); }

      bool trigSun = (dotSun > fDotCut);
      if(trigSun) { trigSun = trigSun && GetSmartPrescaleDecision(smartPrescale); }
      if(trigSun) { trigSun = trigSun && GetAngularPrescaleDecision(zen_sun, azi_sun, zen_trk, azi_trk); }
      	

      if(trigMoon || trigSun){
	auto mint = hlist.front().TDC().val;
	auto maxt = hlist.back().TDC().val;
	for(const DAQHit& it: hlist){
	  mint = std::min(mint, it.TDC().val);
	  maxt = std::max(maxt, it.TDC().val);
	}

	// Add 12.5us padding to each end of the window
	mint -= 800;
	maxt += 800;

	if(trigMoon){
	  ++fNMoonTrigs;
	  td->emplace_back(mint, maxt-mint,
			   daqdataformats::trigID::TRIG_ID_DATA_MOON,
			   fPrescale);
	}
	if(trigSun){
	  ++fNSunTrigs;
	  td->emplace_back(mint, maxt-mint,
			   daqdataformats::trigID::TRIG_ID_DATA_SUN,
			   fPrescale);
	}
      }

    } // end of tracks
    


    const bool ret = !td->empty();
    evt.put(std::move(td));
    return ret;
  }

  void MoonShadow::endJob()
  {
    delete rate_collected_angular;
    std::cout << "=== novaddt::MoonShadow endJob  " << std::endl;
    std::cout << "\tNumber of events:             " << fNEvents    << std::endl;
    std::cout << "\tNumber of tracks:             " << fNTracks    << std::endl;
    std::cout << "\tNumber of sun triggers:       " << fNSunTrigs  << std::endl;
    std::cout << "\tNumber of moon triggers:      " << fNMoonTrigs << std::endl;  
  }
  

  DEFINE_ART_MODULE(MoonShadow)

} // namespace
