////////////////////////////////////////////////////////////////////////
// Class:       NuESelect
// Module Type: filter
// \file:       NuESelect_module.cc
// \brief   	Select Clusters of activity, trigger on EM like Clusters
// \version 	$Id: NuESelect_module.cc,v 1.2 2012-11-19 20:41:15 janzirn Exp $
// \author 	$Author: janzirn $
// \date  	$Date: 2012-11-19 20:41:15 $
//
// Generated at Thu Oct 18 17:27:18 2012 by Jan Zirnstein using artmod
// from art v1_01_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindOneP.h"

#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/Cluster.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTUtilities/func/MathUtil.h"

#include <algorithm>
#include <cmath>
#include <vector>

namespace novaddt {
  class NuESelect;
}

class novaddt::NuESelect : public art::EDFilter {
public:
  explicit NuESelect(fhicl::ParameterSet const & p);
  virtual ~NuESelect();

  virtual bool filter(art::Event & e);

  virtual void beginJob();
  virtual void endJob();

private:

  // Declare member data here.
  std::string _clustlabel;
  std::string _instan;
  double _linelike;
  bool _trigflag;
  unsigned int _prescale;
  unsigned int _tcounts;
};


novaddt::NuESelect::NuESelect(fhicl::ParameterSet const & p)
:
// Initialize member data here.
_clustlabel (p.get< std::string	>("cluster_label")),
_instan  (p.get< std::string    >("instance")),
_linelike (p.get< double	>("LinFitVal")),
_trigflag (false),
_prescale (p.get<unsigned int>("presale",1)),
_tcounts (0)
{
  // Call appropriate Produces<>() functions here.
    produces< std::vector<TriggerDecision>>();
}

novaddt::NuESelect::~NuESelect()
{
  // Clean up dynamic memory and other resources here.
}

void novaddt::NuESelect::beginJob()
{
  // Implementation of optional member function here.
}

bool novaddt::NuESelect::filter(art::Event & e)
{
  // Define product to be put in the event
  std::unique_ptr<std::vector<novaddt::TriggerDecision>> td(new std::vector<novaddt::TriggerDecision>);
 
  // Grab the clusters from the event
  art::Handle<std::vector<novaddt::Cluster>> clusters;
  e.getByLabel(_clustlabel, clusters);

  if(clusters->empty()){
    mf::LogWarning ("No Clusters")<<"No Clusters in the input file";
    return false;
  }

  // Grab the hitlists associated with the clusters
  art::FindOneP<novaddt::HitList> hls(clusters, e, _clustlabel);

  // Time to go through the clusters, C++11 style!
  unsigned int i(0);
  for(auto const& clust: *clusters) {
    art::Ptr<novaddt::HitList> hl_ptr(hls.at(i));
    ++i;

    // Split the hits in x and y to do 2D fitting
    novaddt::HitList hlist(*hl_ptr);
    auto bound = std::partition(hlist.begin(),hlist.end(),CompareDAQHit<View>());

    unsigned int xsize = bound-hlist.begin();
    unsigned int ysize = hlist.end()-bound;

    // Smarter way of vector managment
    std::vector<double> XXpoints;
    std::vector<double> XZpoints;
    std::vector<double> YYpoints;
    std::vector<double> YZpoints;

    XXpoints.reserve(xsize);
    XZpoints.reserve(xsize);
    YYpoints.reserve(ysize);
    YZpoints.reserve(ysize);

    {
      auto it = hlist.begin();
      for(it=hlist.begin();it!=bound;++it){
        XXpoints.emplace_back((*it).Cell().val);
        XZpoints.emplace_back((*it).Plane().val);
      }
      for(it=bound;it!=hlist.end();++it){
        YYpoints.emplace_back((*it).Cell().val);
        YZpoints.emplace_back((*it).Plane().val);
      }
    }

    // Now we can do the actual fitting
    double xslope;
    double xint;
    unsigned int xdof = xsize-1;
    double xsigm=0.0;
    double pred;

    // First the X fit
    if(xsize > 1){
      util::LinFitUnweighted(XXpoints, XZpoints, xslope, xint);

      // Evaluate the goodness of fit
      for(unsigned int xi=0;xi<xsize;++xi){
        pred = xslope*XXpoints[xi] + xint;
        xsigm += sqrt(util::sqr(XZpoints[xi]-pred));
      }
    }
    else xdof = 1;

    // Now the Y fit
    double yslope;
    double yint;
    unsigned int ydof = ysize-1;
    double ysigm=0.0;

    if(ysize > 1){
      util::LinFitUnweighted(YYpoints, YZpoints, yslope, yint);

      // Evaluate the goodness of fit
      for(unsigned int yi=0;yi<ysize;++yi){
        pred = yslope*YYpoints[yi] + yint;
        ysigm += sqrt(util::sqr(YZpoints[yi]-pred));
      }
    }
    else ydof = 1;

    // Let's see if it passes trigger criterion
    double ThreeDFit = sqrt(util::sqr(xsigm/xdof)+util::sqr(ysigm/ydof));
    if(_linelike < ThreeDFit) {
      ++_tcounts;
      if (_tcounts%_prescale == 0) {
        td->emplace_back(clust.EarliestTDC().val, 
			clust.LatestTDC().val - clust.EarliestTDC().val, 
			daqdataformats::trigID::TRIG_ID_DATA_NUE,
			_prescale);
      }
    }
  } // end of clusters

  _trigflag = (td->size() > 0);
  e.put(std::move(td));
  return _trigflag;
} // end filter


void novaddt::NuESelect::endJob()
{
  // Implementation of optional member function here.
}

DEFINE_ART_MODULE(novaddt::NuESelect)
