////////////////////////////////////////////////////////////////////////
// Class:       DDTTrackMerge
// Module Type: filter
// File:        DDTTrackMerge_module.cc
// Author: Zukai Wang
////////////////////////////////////////////////////////////////////////
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Persistency/Common/Assns.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/Track.h"
#include "DDTBaseDataProducts/Track3D.h"
#include "DDTBaseDataProducts/HitList.h"

#include "DAQChannelMap/DAQChannelMap.h"
#include "DDTUtilities/AssociationUtil.h"

#include <memory>
#include <vector>

//------------------------------------------------------------------------
namespace novaddt {
  class DDTTrackMerge;
}
//------------------------------------------------------------------------
class novaddt::DDTTrackMerge : public art::EDFilter {
public:
  explicit DDTTrackMerge(fhicl::ParameterSet const & p);
  virtual ~DDTTrackMerge();
  bool    filter(art::Event & e) override;
  void    endJob() override;

  void    simpleMerge(std::vector< art::Ptr<novaddt::Track> > tracks, 
		      std::unique_ptr< std::vector<novaddt::Track3D> > & merged_tracks, 
		      std::unique_ptr< std::vector<novaddt::HitList> > & hitListcol,
		      art::FindOneP<novaddt::HitList> fohl);

private:
  std::string  _trackModuleLabel;         ///< label of module making the Tracks
  std::string  _trackInstanceLabel;       ///< instance label making the Tracks
  std::string  _sliceModuleLabel;         ///< label of module making the HitList
  std::string  _sliceInstanceLabel;       ///< instance label making the HitList
  std::string  _trackToSliceInstanceLabel;///< instance of track to slice association produced by track module

  float        _Zdiff;

  int          _nSlices = 0;              ///< number of slices seen
  int          _nXTracks = 0;              ///< number of x tracks seen
  int          _nYTracks = 0;              ///< number of y tracks seen

  int          _nMergedTracks = 0;        ///< number of merged tracks produced
};
//------------------------------------------------------------------------
novaddt::DDTTrackMerge::DDTTrackMerge(fhicl::ParameterSet const & p)
:   _trackModuleLabel           (p.get<std::string >("TrackModuleLabel"      ))
  , _trackInstanceLabel         (p.get<std::string >("TrackInstanceLabel", ""))
  , _sliceModuleLabel           (p.get<std::string >("SliceModuleLabel"      ))
  , _sliceInstanceLabel         (p.get<std::string >("SliceInstanceLabel", ""))
  , _trackToSliceInstanceLabel  (p.get<std::string >("TrackToSliceInstanceLabel", ""))
  , _Zdiff                      (p.get<float>       ("Z_Diff"))

{
    std::cout << "=== novaddt::DDTTrackMerge instantiate" << std::endl;
    std::cout << "Here is the configuration"<<std::endl;
    std::cout << "\t TrackModuleLabel:          " << _trackModuleLabel          << std::endl;
    std::cout << "\t TrackInstanceLabel:        " << _trackInstanceLabel        << std::endl;
    std::cout << "\t SliceModuleLabel:          " << _sliceModuleLabel          << std::endl;
    std::cout << "\t SliceInstanceLabel:        " << _sliceInstanceLabel        << std::endl;
    std::cout << "\t TrackToSliceInstanceLabel: " << _trackToSliceInstanceLabel << std::endl;

    std::cout << "\t Zdiff:                     " << _Zdiff                     << std::endl;
 
    produces< std::vector<novaddt::Track3D> >();
    produces< std::vector<novaddt::HitList> >();
    
    produces< art::Assns<novaddt::Track3D, novaddt::HitList> >();
    produces< art::Assns<novaddt::Track3D, novaddt::HitList> >(_trackToSliceInstanceLabel);
}
//------------------------------------------------------------------------
novaddt::DDTTrackMerge::~DDTTrackMerge()
{
    // Clean up dynamic memory and other resources here.
}

//------------------------------------------------------------------------
bool novaddt::DDTTrackMerge::filter(art::Event & event)
{
    // Implementation of required member function here.
    std::cout << "=== novaddt::DDTTrackMerge filter. Event: " 
              << event.id().event() 
              << std::endl;
    
    // make the final data products
    std::unique_ptr< std::vector<novaddt::Track3D> > merged_tracks    (new std::vector<novaddt::Track3D>);
    std::unique_ptr< std::vector<novaddt::HitList> > merged_hit_lists (new std::vector<novaddt::HitList>);
    std::unique_ptr< art::Assns<novaddt::Track3D,   novaddt::HitList> > assn(new art::Assns<novaddt::Track3D, novaddt::HitList>);
    std::unique_ptr< art::Assns<novaddt::Track3D,   novaddt::HitList> > assn_b(new art::Assns<novaddt::Track3D, novaddt::HitList>);
    
    // get the slices
    art::Handle< std::vector<novaddt::HitList> > slices;
    event.getByLabel(_sliceModuleLabel, _sliceInstanceLabel, slices);

    _nSlices+=slices->size();
    
    //Now make it an art::PtrVector so we can use associations later
    art::PtrVector<novaddt::HitList> slicelist;
    for(unsigned int i = 0; i < slices->size(); ++i){
        art::Ptr<novaddt::HitList>slice_ptr(slices,i);
        slicelist.push_back(slice_ptr);
    }
    
    const std::string the_slices = _trackModuleLabel+":"+_trackToSliceInstanceLabel;
    art::FindManyP<novaddt::Track> find_tracks(slices, event, the_slices);
    
    // loop over slices
    for(size_t i_slice = 0; i_slice < slices->size(); ++i_slice){
      // get the hit list
        novaddt::HitList hit_list = slices->at(i_slice);
        // find the tracks for this hit list
        std::vector< art::Ptr<novaddt::Track> > tracks = find_tracks.at(i_slice);

        if (!tracks.size()) continue;

        // get the hit list for each track
        art::FindOneP<novaddt::HitList> fohl(tracks, event, _trackModuleLabel);
 
	unsigned int current_size = merged_tracks->size();
        // merge tracks
	simpleMerge(tracks, merged_tracks, merged_hit_lists, fohl);               

        // the number of merged tracks should ALWAYS be equal to the number of hit lists
        assert(merged_tracks->size()==merged_hit_lists->size());
         
        // loop over new tracks and hit lists, save final products and associations
        for (unsigned int i = current_size; i < merged_tracks->size(); ++i){
	  // association between this slice and each track
	  util::CreateAssn(*this, event, *merged_tracks, slicelist[i_slice], *assn_b, i);
	  // association between each track and it's hits
	  util::CreateAssn(*this, event, *merged_tracks, *merged_hit_lists, *assn, i, i+1, i);
        }        
    } // end of loop on slices
    
    // store the merged tracks

    event.put(std::move(merged_tracks));
    event.put(std::move(merged_hit_lists));
    event.put(std::move(assn));
    event.put(std::move(assn_b), _trackToSliceInstanceLabel);
    return true;
}

//------------------------------------------------------------------------
void novaddt::DDTTrackMerge::simpleMerge(std::vector< art::Ptr<novaddt::Track> > tracks,
                                         std::unique_ptr< std::vector<novaddt::Track3D> > & merged_tracks,
                                         std::unique_ptr< std::vector<novaddt::HitList> > & merged_hit_lists,
                                         art::FindOneP<novaddt::HitList> fohl)
{
 
  //First, we need to classify the tracks in the same view:
  std::vector<art::Ptr<novaddt::Track> > Xtracks;
  std::vector<art::Ptr<novaddt::Track> > Ytracks;
  std::vector<bool>  y_used;
  std::vector<int>   x_Tridx, y_Tridx;

  for(size_t i_track = 0; i_track < tracks.size(); ++i_track){
    if (tracks[i_track]->View() == daqchannelmap::X_VIEW) {
      Xtracks.emplace_back(tracks[i_track]);

      x_Tridx.emplace_back(i_track);
    }
    else {
      Ytracks.emplace_back(tracks[i_track]);
      y_used.push_back(0);
      y_Tridx.emplace_back(i_track);
    }
  }
    
  _nXTracks+=Xtracks.size();
  _nYTracks+=Ytracks.size();

    
  if ( (Xtracks.size() == 0) || (Ytracks.size() == 0) ) return;

  //Start Merging:
  novaddt::HitList merged_hit_list;

  for (size_t i=0; i!= Xtracks.size();++i) {
    float x_low_z  = std::min(Xtracks[i]->StartZ(),Xtracks[i]->EndZ());
    float x_high_z = std::max(Xtracks[i]->StartZ(),Xtracks[i]->EndZ());
    for (size_t j=0;j!=Ytracks.size();++j) {
      if (y_used[j]) continue;
      float y_low_z  = std::min(Ytracks[j]->StartZ(),Ytracks[j]->EndZ());
      float y_high_z = std::max(Ytracks[j]->StartZ(),Ytracks[j]->EndZ());
      if (fabs(x_low_z-y_low_z)< _Zdiff && fabs(x_high_z-y_high_z)<_Zdiff) {
	y_used[j] = true;

	TVector3 start;
	TVector3 end;
	float x0,y0,x1,y1;
	x0=0;
	y0=0;
	x1=0;
	y1=0;
	float xz0 = 9999;
	float yz0 = 9999;
	float xz1 = 0;
	float yz1 = 0;


	// save merged hit list
	novaddt::HitList merged_hit_list;

	for (auto const& hit : *fohl.at(x_Tridx[i]) ) {
	  merged_hit_list.emplace_back(hit);	  
	  if (xz0 > hit.Plane().val) {
	    x0  = hit.Cell().val;
	    xz0 = hit.Plane().val;
	  }
	  if (xz1 < hit.Plane().val) {
	    x1  = hit.Cell().val;
	    xz1 = hit.Plane().val;
	  }
	  
	}

	for (auto const& hit : *fohl.at(y_Tridx[j]) ) {
	  merged_hit_list.emplace_back(hit);	  
	  if (yz0 > hit.Plane().val) {
	    y0  = hit.Cell().val;
	    yz0 = hit.Plane().val;
	  }
	  if (yz1 < hit.Plane().val) {
	    y1  = hit.Cell().val;
	    yz1 = hit.Plane().val;
	  }
	  
	}            
	merged_hit_lists->push_back(merged_hit_list);
 
	++_nMergedTracks;        

	start.SetXYZ(x0,
		     y0,
		     std::min(xz0,yz0));
	end.SetXYZ(  x1,
		     y1,
		     std::max(xz1,yz1));

	novaddt::Track3D merged_track = Track3D(true, 2, start, end);	
	merged_tracks->push_back(merged_track);    
	break;     
      }
    }
  }

}

//------------------------------------------------------------------------
void novaddt::DDTTrackMerge::endJob()
{
    std::cout << "=== novaddt::DDTTrackMerge endJob" << std::endl;
    std::cout << "\tNumber of slices seen:        " << _nSlices           << std::endl;
    std::cout << "\tNumber of x tracks seen:      " << _nXTracks          << std::endl;
    std::cout << "\tNumber of y tracks seen:      " << _nYTracks          << std::endl;
    std::cout << "\tNumber of merged tracks made: " << _nMergedTracks     << std::endl;
}
//------------------------------------------------------------------------
DEFINE_ART_MODULE(novaddt::DDTTrackMerge)
