//////////////////////////////////////////////////////////////////////
// Class:       Slicer3D
// Module Type: producer
// File:        Slicer3D_module.cc
//
//////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"

#include "DAQChannelMap/DAQChannelMap.h"

//////////////////////////////////////////////////////////////////////
namespace novaddt {
    class Slicer3D;
}
//////////////////////////////////////////////////////////////////////
class novaddt::Slicer3D : public art::EDFilter {
public:
  explicit Slicer3D(fhicl::ParameterSet const & p);
  virtual ~Slicer3D();
  virtual bool filter(art::Event & event);
  void endJob() override;
private:
  // Declare member data
  std::string     _InputLabel;
  std::string     _InputInstance;
  unsigned        fMaxCellGap;
  unsigned        fMinHit;

  // Functions
  void makeSlice(HitList & hits, std::vector<HitList> & hitlistcol);
  // Counters
  int _nEvents        = 0;
  int _nInputSlices   = 0;
  int _nOutputSlices  = 0;
  int _nHitsIn        = 0;
  int _nHitsOut       = 0;
};
//////////////////////////////////////////////////////////////////////
novaddt::Slicer3D::Slicer3D(fhicl::ParameterSet const & p)
:
_InputLabel     (p.get< std::string >("InputLabel"   )),
_InputInstance  (p.get< std::string >("InputInstance")),
fMaxCellGap     (p.get< int         >("MaxCellGap"   )),
fMinHit         (p.get< int         >("MinHit"       ))

// Initialize member data
{
    // Call appropriate Produces<>() functions
    produces<std::vector<HitList> >("Zuk3DSlicer");
}
//////////////////////////////////////////////////////////////////////
novaddt::Slicer3D::~Slicer3D()
{
    // Clean up dynamic memory and other resources here.
}
//////////////////////////////////////////////////////////////////////
bool  novaddt::Slicer3D::filter(art::Event & event)
{
    _nEvents++;
    
    // Array of hitlists from event (input)
    art::Handle<std::vector<HitList> > hit_lists; // from HitList
    event.getByLabel(_InputLabel, _InputInstance, hit_lists);
    
    // Array of filtered hitlists (output)
    std::unique_ptr< std::vector<HitList> >hitList_col(new std::vector<HitList>);

    // Loop over grouped hitlists (slices) to extract hitlists
    for(unsigned int i=0; i < hit_lists->size(); ++i){
      HitList xHits,yHits;

      _nInputSlices++;
      HitList const &hits = hit_lists->at(i);
      _nHitsIn += hits.size(); // # of hits in input .root file

      for (auto hit = hits.begin(); hit != hits.end(); ++hit) {

	//Check View:
	if  (hit->View().val == daqchannelmap::X_VIEW) 
	  xHits.emplace_back(*hit);
	else yHits.emplace_back(*hit);
      }

      if (xHits.size()>=fMinHit) makeSlice(xHits,*hitList_col);
      if (yHits.size()>=fMinHit) makeSlice(yHits,*hitList_col);

    } // End of hitlists loop
    
    // Add filtered hitlists to event
    _nOutputSlices+=hitList_col->size();
    event.put(std::move(hitList_col),"Zuk3DSlicer");
    
    return true;
}
//////////////////////////////////////////////////////////////////////
void  novaddt::Slicer3D::makeSlice(HitList & hits, std::vector<HitList> & hitlistcol)
{
  //Sort this hitlist by cell (it is already sliced by plane and time)
  std::sort(hits.begin(),hits.end(),CompareDAQHit<Cell>());

  auto itStart = hits.begin();
  auto itEnd   = hits.begin();

  while (itEnd != hits.end()) {
    const unsigned endCell = (*itEnd).Cell().val;
    ++itEnd;
    const unsigned nextCell = (*itEnd).Cell().val;
      
    if (nextCell - endCell > fMaxCellGap) {
      if (itEnd - itStart > fMinHit-2) {
	HitList frag(itStart, itEnd);
	_nHitsOut+=frag.size();
	hitlistcol.push_back(frag);
      }
      itStart = itEnd;
    }      
  }

  if (itEnd-itStart > fMinHit-2) {
    HitList frag(itStart, itEnd);
    _nHitsOut+=frag.size();
    hitlistcol.push_back(frag);
  }


  return;
}
//////////////////////////////////////////////////////////////////////
void novaddt::Slicer3D::endJob()
{
    std::cout << "=== novaddt::Slicer3D end ===" << std::endl;
    std::cout << "\t# of events:             " << _nEvents       << std::endl;
    std::cout << "\t# of slices input:     " << _nInputSlices  << std::endl;
    std::cout << "\t# of slices output:    " << _nOutputSlices << std::endl;
    std::cout << "\t# of hits input:         " << _nHitsIn << std::endl;
    std::cout << "\t# of hits output:        " << _nHitsOut << std::endl;
}
//////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(novaddt::Slicer3D)
