#include "SlowMonopoleTrigger/Track.h"

#include "SlowMonopoleTrigger/Constants.h"



novaddt::smt::Track::Track
(novaddt::DAQHit const& start, novaddt::DAQHit const& end)
  : start_(start), end_(end)
{
  dp_ = static_cast<double>(end_.Plane().val) -
        static_cast<double>(start_.Plane().val);
  dc_ = static_cast<double>(end_.Cell().val) -
        static_cast<double>(start_.Cell().val);
  
  dt_ = static_cast<int64_t>(end_.TDC().val) -
        static_cast<int64_t>(start_.TDC().val);

  distance_ = calculate_distance();
  velocity_ = calculate_velocity();
  beta_     = velocity_ / Constants::SPEED_OF_LIGHT;

  calculate_slopes();
}



double novaddt::smt::Track::calculate_distance() const
{
  double dx = Constants::CELL_WIDTH  * dc_;
  double dz = Constants::PLANE_WIDTH * dp_;

  return sqrt(dx*dx + dz*dz);
}



double novaddt::smt::Track::calculate_velocity() const
{
   double dt_in_nanoseconds = Constants::TDC_TICK * dt_;

   return distance_ / dt_in_nanoseconds;
}



void novaddt::smt::Track::calculate_slopes()
{
  if (dp_ == 0)
  {
    slope_plane_is_valid_ = false;
  } else {
    slope_plane_is_valid_ = true;
    slope_cell_plane_ = dc_ / dp_;
    slope_time_plane_ = dt_ / dp_;
  }

  if (dt_ == 0)
  {
    slope_time_is_valid_ = false;
  } else {
    slope_time_is_valid_ = true;
    slope_cell_time_  = dc_ / dt_;
    slope_plane_time_ = dp_ / dt_;
  }

  if (dc_ == 0)
  {
    slope_cell_is_valid_ = false;
  } else {
    slope_cell_is_valid_ = true;
    slope_time_cell_ = dt_ / dc_;
  }
}



double novaddt::smt::Track::distance() const
{
  return distance_;
}



double novaddt::smt::Track::time() const
{
  return dt_;
}



double novaddt::smt::Track::velocity() const
{
  return velocity_;
}



double novaddt::smt::Track::beta() const
{
  return beta_;
}



bool novaddt::smt::Track::slope_plane_is_valid() const
{
  return slope_plane_is_valid_;
}



bool novaddt::smt::Track::slope_time_is_valid() const
{
  return slope_time_is_valid_;
}



bool novaddt::smt::Track::slope_cell_is_valid() const
{
  return slope_cell_is_valid_;
}



double novaddt::smt::Track::slope_cell_plane() const
{
  if (!slope_plane_is_valid())
  {
    std::cerr << "\n\tThe slope is invalid.\n\n"
	      << "Please check Track::slope_is_valid() "
	      << "before accessing Track::slope()." << std::endl;
    assert(false);
  }

  return slope_cell_plane_;
}



double novaddt::smt::Track::slope_time_plane() const
{
  if (!slope_plane_is_valid())
  {
    std::cerr << "\n\tThe slope is invalid.\n\n"
	      << "Please check Track::slope_is_valid() "
	      << "before accessing Track::slope()." << std::endl;
    assert(false);
  }

  return slope_time_plane_;
}



double novaddt::smt::Track::slope_cell_time() const
{
  if (!slope_time_is_valid())
  {
    std::cerr << "\n\tThe slope is invalid.\n\n"
	      << "Please check Track::slope_is_valid() "
	      << "before accessing Track::slope()." << std::endl;
    assert(false);
  }

  return slope_cell_time_;
}



double novaddt::smt::Track::slope_plane_time() const
{
  if (!slope_time_is_valid())
  {
    std::cerr << "\n\tThe slope is invalid.\n\n"
	      << "Please check Track::slope_is_valid() "
	      << "before accessing Track::slope()." << std::endl;
    assert(false);
  }

  return slope_plane_time_;
}



double novaddt::smt::Track::slope_time_cell() const
{
  if (!slope_cell_is_valid())
  {
    std::cerr << "\n\tThe slope is invalid.\n\n"
	      << "Please check Track::slope_is_valid() "
	      << "before accessing Track::slope()." << std::endl;
    assert(false);
  }

  return slope_time_cell_;
}



novaddt::DAQHit novaddt::smt::Track::start() const
{
  return start_;
}



novaddt::DAQHit novaddt::smt::Track::end() const
{
  return end_;
}



unsigned novaddt::smt::Track::min_plane() const
{
  return std::min(start_.Plane().val, end_.Plane().val);
}



unsigned novaddt::smt::Track::max_plane() const
{
  return std::max(start_.Plane().val, end_.Plane().val);
}



int novaddt::smt::Track::plane_difference() const
{
  return dp_;
}



unsigned novaddt::smt::Track::min_cell() const
{
  return std::min(start_.Cell().val, end_.Cell().val);
}



unsigned novaddt::smt::Track::max_cell() const
{
  return std::max(start_.Cell().val, end_.Cell().val);
}



int novaddt::smt::Track::cell_difference() const
{
  return dc_;
}
